Package: TidyPanel
Title: Universal Messy Panel Data Cleaner
Version: 0.1.2
Authors@R: person("Tony", "Lu", email = "xulunt123@gmail.com", role = c("aut", "cre"))
Description: A robust toolkit designed to standardize and clean complex tabular data
    from commercial enterprise systems, healthcare records, logistics software, and HR
    databases. Features include intelligent regex parsing for domain-specific noise
    (currencies, percentages), gap-based block clustering, and automated messy table
    resolution. Methods draw on tidy data principles described in Wickham (2014)
    <doi:10.18637/jss.v059.i10> and the 'readxl' parsing infrastructure described in
    Wickham & Bryan (2023) <https://readxl.tidyverse.org>.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: dplyr, stringr, readxl
Suggests: knitr, rmarkdown, testthat, writexl
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-05-06 03:35:31 UTC; TonyL
Author: Tony Lu [aut, cre]
Maintainer: Tony Lu <xulunt123@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-11 19:00:09 UTC
Built: R 4.6.0; ; 2026-05-11 19:37:25 UTC; unix

# Generated by roxygen2: do not edit by hand

export(clean_variable_names)
export(infer_data_types)
export(normalize_units)
export(read_messy_panel)
importFrom(readxl,excel_sheets)
importFrom(readxl,read_excel)
importFrom(stringr,str_extract)
importFrom(stringr,str_remove)
importFrom(stringr,str_remove_all)
importFrom(stringr,str_replace)
importFrom(stringr,str_squish)
importFrom(stringr,str_to_lower)
importFrom(stringr,str_trim)

# TidyPanel <img src="man/figures/logo.png" align="right" height="139" />

<!-- badges: start -->
<!-- badges: end -->

**TidyPanel** is an industrial-grade parser designed to extract clean, standardized data frames from heavily malformed, human-readable Excel reports. If you have ever struggled to parse financial statements, ERP exports, or complex tables with N-dimensional headers, decoy rows, and embedded subtotals, `TidyPanel` is built for you.

## Installation

You can install the development version of TidyPanel from GitHub with:

``` r
# install.packages("devtools")
devtools::install_github("TonyL/TidyPanel")
```

## Why TidyPanel?

Real-world commercial data is rarely tidy. It contains multi-line merged headers, embedded subtotal rows, empty "ghost" columns, and categorical hierarchies defined purely by visual indentation (e.g., in a Profit & Loss statement).

`TidyPanel` uses a multi-phase heuristic engine to:
1. **Bypass Decoy Rows**: Skips irrelevant metadata at the top of the sheet.
2. **N-Dimensional Header Stitching**: Intelligently identifies multiline headers and forward-fills merged cells to create flat, readable column names.
3. **Indentation Hierarchy Extraction**: Uses leading spaces to identify parent-child categorical relationships and creates a `parent_category` column.
4. **Smart Data Amputation**: Automatically removes decorative page breaks, mid-table subtotals, and ghost aggregate rows.
5. **Auto Pivot**: Detects temporal columns (e.g., Q1, 2021, FY23) and pivots them into a tidy, long format.
6. **Semantic Cleaners**: Automatically detects and normalizes accounting dashes, non-standard scientific notation, and financial multipliers (e.g., "1.5M" -> 1500000).

## 100% Transparency: The Audit Log

Data parsing should never be a "black box". TidyPanel features a unique **Audit Log** that explicitly records every transformation applied to your data. By setting `return_audit = TRUE`, you get both the cleaned data and a precise ledger of what was modified.

``` r
library(TidyPanel)

# Read data and get an audit trail
result <- read_messy_panel("data_raw/financial_report.xlsx", return_audit = TRUE)

# View the audit trail
print(result$audit)
```
```text
=== Algorithm Modification Audit Log ===
                        Operation Count
1             Decoy Rows Bypassed     5
2 Indentation Hierarchy Extracted     3
3       Ghost Bottom Rows Dropped     4
```

## License

MIT © Tony Lu

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----basic_example------------------------------------------------------------
library(TidyPanel)

# Create a minimal example with financial multiplier notation
tmp <- tempfile(fileext = ".xlsx")
df_raw <- data.frame(
  Category  = c("Revenue", "Cost"),
  `FY2022`  = c("1.5M", "800k"),
  `FY2023`  = c("2.0M", "950k"),
  check.names = FALSE
)
writexl::write_xlsx(df_raw, tmp)

# Parse and auto-pivot the wide table to long format
result <- read_messy_panel(tmp, auto_pivot = TRUE)
head(result)

unlink(tmp)

## ----vars_example-------------------------------------------------------------
# Demonstrate variable name standardization
df_messy <- data.frame(
  `GVKEY`            = c("001", "002"),
  `Total Revenue ($)` = c(100, 200),
  `My Custom Col!`   = c("A", "B"),
  check.names = FALSE
)

df_clean <- clean_variable_names(df_messy)
colnames(df_clean)

## ----audit_example------------------------------------------------------------
# Create a table with a mid-table subtotal row (to be amputated)
tmp2 <- tempfile(fileext = ".xlsx")
df_with_subtotal <- data.frame(
  Item    = c("Product A", "Product B", "Total", "Product C"),
  Revenue = c("100", "200", "300", "150")
)
writexl::write_xlsx(df_with_subtotal, tmp2)

# Parse with audit log enabled
result <- read_messy_panel(tmp2, return_audit = TRUE)

# View the cleaned data (Total row should be amputated)
head(result$data)

# View what the engine did
print(result$audit)

unlink(tmp2)

#' Robust Parsing and Extraction of Messy Excel Panel Data
#'
#' @description
#' `read_messy_panel()` is an industrial-grade parser designed to extract clean, standardized data frames 
#' from heavily malformed, human-readable Excel reports (e.g., financial statements, ERP exports). 
#' It automatically bypasses decoy rows, stitches N-dimensional hierarchical headers, extracts structural 
#' indentation hierarchies (parent-child relationships), amputates embedded subtotals, and standardizes 
#' financial/scientific numbers.
#'
#' @param file_path Character string. Path to the Excel file.
#' @param sheet Optional sheet name or index. If `NULL` (the default), it auto-discovers the first valid data panel across all sheets. If `"ALL"`, parses and merges all sheets.
#' @param na_strings Character vector. Strings to interpret as missing values. Supports complex missing-value lexicons.
#' @param clean_vars Logical. If `TRUE` (default), standardizes variable names to snake_case using `clean_variable_names()`.
#' @param auto_pivot Logical. If `TRUE`, attempts to reshape wide temporal columns (e.g., FY2021, Q1_2022) into a long format (`time_period`, `value`).
#' @param return_audit Logical. If `TRUE`, returns a list containing `$data` (the cleaned data frame) and `$audit` (a detailed log of all algorithmic modifications made).
#'
#' @return If `return_audit = FALSE`, a cleaned and standardized `data.frame`. 
#' If `return_audit = TRUE`, a named list containing:
#' \item{data}{The cleaned `data.frame`.}
#' \item{audit}{A `data.frame` detailing exactly what transformations, truncations, or imputations were applied.}
#'
#' @examples
#' # Toy example: create a small in-memory Excel file and parse it
#' tmp <- tempfile(fileext = ".xlsx")
#' df_raw <- data.frame(
#'   Category = c("Revenue", "Cost", "Total"),
#'   `2022` = c("1.2M", "800k", "2.0M"),
#'   `2023` = c("1.5M", "900k", "2.4M"),
#'   check.names = FALSE
#' )
#' writexl::write_xlsx(df_raw, tmp)
#' result <- read_messy_panel(tmp, auto_pivot = TRUE)
#' head(result)
#' unlink(tmp)
#'
#' @export
#' @importFrom readxl read_excel excel_sheets
#' @importFrom stringr str_trim str_replace str_remove_all str_squish str_extract
read_messy_panel <- function(file_path, sheet = NULL, na_strings = c("", "NA", "#N/A", "NULL", "S", "D", "ND", "N/A", "*", "**", "***", ".", "x", "c", "s", "z", "#VALUE!", "#REF!", "#DIV/0!", "#NUM!", "#NAME?", "none", "NR", "--", "---", "n.a.", "N.A.", "n/a", "Not Applicable"), clean_vars = TRUE, auto_pivot = FALSE, return_audit = FALSE) {
  
  audit_log <- list()
  
  if (is.character(sheet) && length(sheet) == 1 && toupper(sheet) == "ALL") {
      sheets <- readxl::excel_sheets(file_path)
      all_data <- list()
      all_audits <- list()
      
      for (s in sheets) {
          res <- tryCatch({
              read_messy_panel(file_path, sheet = s, na_strings = na_strings, clean_vars = clean_vars, auto_pivot = auto_pivot, return_audit = TRUE)
          }, error = function(e) list(error = e$message))
          
          if ("data" %in% names(res)) {
              df_s <- res$data
              if (nrow(df_s) > 0) {
                  df_s$source_sheet_name <- s
                  df_s <- df_s[, c("source_sheet_name", setdiff(colnames(df_s), "source_sheet_name")), drop = FALSE]
              }
              all_data[[s]] <- df_s
              all_audits[[s]] <- res$audit
          } else {
              all_audits[[s]] <- list(error = res$error)
          }
      }
      
      if (length(all_data) == 0) {
          stop("Could not parse any sheet in the workbook.")
      }
      
      master_df <- dplyr::bind_rows(all_data)
      
      if (return_audit) {
          return(list(success = TRUE, data = master_df, audit = all_audits))
      } else {
          return(master_df)
      }
  }
  
  sheets <- readxl::excel_sheets(file_path)
  sheets_to_try <- if (!is.null(sheet)) sheet else sheets
  
  is_numeric_like <- function(x) {
    if (is.na(x) || x %in% na_strings) return(TRUE) 
    clean_x <- stringr::str_remove_all(x, intToUtf8(160))
    if (stringr::str_trim(clean_x) %in% c("-", "\u2013", "\u2014")) return(TRUE)
    if (grepl("^[-\u2014\u2013\u2014]+$", stringr::str_trim(clean_x))) return(TRUE)
    if (grepl("(?i)^(Q[1-4]|H[1-2]|FY[0-9]+)$", stringr::str_trim(clean_x))) return(FALSE)
    
    # Phase 5: Non-standard scientific notation
    clean_x <- stringr::str_replace_all(clean_x, "(?i)\\s*[x\\*]\\s*10\\^([\\-\\+]?[0-9]+)", "E\\1")
    clean_x <- stringr::str_replace(clean_x, "^\\s*\\((.*)\\)\\s*$", "-\\1")
    clean_x <- stringr::str_replace(clean_x, "^-\\s+", "-")
    clean_x <- stringr::str_replace(clean_x, "^\\s*([0-9.,\\s]+?)\\s*-$", "-\\1")
    clean_x <- stringr::str_remove_all(clean_x, "[\\$\u20ac\u00a3\u00a5%\u4e07\u4ebf\u5143\u5343]")
    
    clean_x <- stringr::str_replace_all(clean_x, "(?<=\\d)[\\s\\u00A0'](?=\\d)", "")
    has_euro_decimal <- grepl(",[0-9]{1,2}[^0-9]*$", clean_x)
    if (!is.na(has_euro_decimal) && has_euro_decimal) {
        clean_x <- stringr::str_replace(stringr::str_remove_all(clean_x, "\\."), ",", ".")
    }
    clean_x <- stringr::str_remove_all(clean_x, ",")
    
    clean_x <- stringr::str_replace(clean_x, "\\s*\\*+\\s*$", "")
    clean_x <- stringr::str_replace(clean_x, "\\s*[\\(\\[].*?[\\)\\]]\\s*$", "")
    clean_x <- stringr::str_replace(clean_x, "\\s*[A-Za-z\u4e00-\u9fa5]+\\s*$", "")
    clean_x <- stringr::str_replace(clean_x, "^\\s*[A-Za-z\u4e00-\u9fa5]+\\s*", "")
    
    !is.na(suppressWarnings(as.numeric(clean_x)))
  }
  
  last_error <- "Could not parse any sheet."
  
  for (s in sheets_to_try) {
    res <- tryCatch({
      raw_data <- suppressMessages(readxl::read_excel(file_path, sheet = s, col_names = FALSE, .name_repair = "minimal", trim_ws = FALSE))
      if (nrow(raw_data) == 0) stop("Empty sheet")
      
      raw_mat <- as.matrix(raw_data)
      
      num_counts <- apply(raw_mat, 1, function(row) {
        sum(vapply(row, is_numeric_like, logical(1)) & !is.na(row) & row != "")
      })
      
      is_data_row <- num_counts >= 1
      true_runs_indices <- which(is_data_row == TRUE)
      
      if (length(true_runs_indices) == 0) {
        stop("Could not detect any numeric panel data block.")
      }
      
      gaps <- diff(true_runs_indices)
      block_boundaries <- which(gaps > 5)
      
      if (length(block_boundaries) == 0) {
        blocks <- list(true_runs_indices)
      } else {
        blocks <- list()
        start_idx <- 1
        for (b in block_boundaries) {
          blocks <- append(blocks, list(true_runs_indices[start_idx:b]))
          start_idx <- b + 1
        }
        blocks <- append(blocks, list(true_runs_indices[start_idx:length(true_runs_indices)]))
      }
      
      block_lengths <- vapply(blocks, length, integer(1))
      main_block <- blocks[[which.max(block_lengths)]]
      
      start_data_row <- min(main_block)
      end_data_row <- max(main_block)
      
      main_block_counts <- num_counts[start_data_row:end_data_row]
      mode_count <- as.numeric(names(sort(table(main_block_counts), decreasing = TRUE)[1]))
      
      density_threshold <- max(1, floor(mode_count * 0.2))
      
      valid_in_block <- which(main_block_counts >= density_threshold)
      if (length(valid_in_block) == 0) {
          true_start <- start_data_row
          true_end <- end_data_row
      } else {
          true_start <- start_data_row
          true_end <- start_data_row + max(valid_in_block) - 1
      }
      
      while (true_start <= end_data_row) {
          if (num_counts[true_start] < density_threshold) {
              true_start <- true_start + 1
              next
          }
          
          row_vals <- raw_mat[true_start, ]
          num_vals <- suppressWarnings(as.numeric(row_vals[!is.na(row_vals) & row_vals != ""]))
          valid_nums <- num_vals[!is.na(num_vals)]
          is_year_header <- length(valid_nums) >= 2 && all(valid_nums >= 1900 & valid_nums <= 2100)
          
          if (is_year_header) {
              true_start <- true_start + 1
              next
          }
          
          non_empty_current <- sum(!is.na(row_vals) & row_vals != "")
          non_empty_above <- if (true_start > 1) sum(!is.na(raw_mat[true_start-1, ]) & raw_mat[true_start-1, ] != "") else 0
          
          if (true_start > 1 && non_empty_above > 0 && (non_empty_above < non_empty_current * 0.3)) {
              true_start <- true_start + 1
              next
          }
          
          break
      }
      
      if (true_start > end_data_row) true_start <- end_data_row
      
      header_row_index <- max(1, true_start - 1)
      audit_log[["Decoy Rows Bypassed"]] <- header_row_index - 1
      
      if (header_row_index == true_start) {
          true_start <- true_start + 1
      }
      # Helper: detect rows that are pure noise (random-looking uppercase/lowercase strings
      # with no semantic value, e.g. "xdasdad", "WEDEWADAW"). Such rows should NOT be
      # concatenated onto real column names.
      is_noise_header_row <- function(row_vals) {
          non_empty <- row_vals[!is.na(row_vals) & stringr::str_trim(row_vals) != ""]
          if (length(non_empty) == 0) return(FALSE)
          # A row is "noise" if every non-empty cell:
          #   (a) contains only letters (no spaces, digits, underscores, or punctuation), AND
          #   (b) looks random: either all-uppercase with 5+ chars, or a mixed-case string
          #       whose character bigram entropy suggests it has no real word structure.
          looks_random <- vapply(non_empty, function(x) {
              x <- stringr::str_trim(x)
              # Must be letters-only (no spaces/digits/special chars)
              if (!grepl("^[A-Za-z]+$", x)) return(FALSE)
              # Short common words (<=4 chars) are NOT noise (e.g. "ID", "Age")
              if (nchar(x) <= 4) return(FALSE)
              # If it looks like a known word prefix or is title-cased, not noise
              if (grepl("^[A-Z][a-z]", x)) return(FALSE)
              # Otherwise treat as noise (all-caps random string like "WEDEWADAW")
              TRUE
          }, logical(1))
          # Only flag as noise if ALL non-empty cells in this row are random-looking
          all(looks_random)
      }

      header_rows_list <- list()
      for (r in 1:header_row_index) {
          row_vals <- raw_mat[r, ]
          non_empty_count <- sum(!is.na(row_vals) & stringr::str_trim(row_vals) != "")
          if (non_empty_count > 1 || r == header_row_index) {
              # Skip rows that are pure noise garbage
              if (r != header_row_index && is_noise_header_row(row_vals)) {
                  audit_log[["Noise Header Rows Discarded"]] <-
                      c(audit_log[["Noise Header Rows Discarded"]], r)
                  next
              }
              header_rows_list <- append(header_rows_list, list(row_vals))
          }
      }
      
      if (length(header_rows_list) > 0) {
          for (i in seq_along(header_rows_list)) {
              h_row <- header_rows_list[[i]]
              if (length(unique(h_row[!is.na(h_row) & stringr::str_trim(h_row) != ""])) >= 1) {
                  for (j in 2:length(h_row)) {
                      if ((is.na(h_row[j]) || stringr::str_trim(h_row[j]) == "") && 
                          !is.na(h_row[j-1]) && stringr::str_trim(h_row[j-1]) != "") {
                          h_row[j] <- h_row[j-1]
                      }
                  }
                  header_rows_list[[i]] <- h_row
              }
          }
          
          headers <- rep("", length(header_rows_list[[1]]))
          for (i in seq_along(header_rows_list)) {
              h_row <- header_rows_list[[i]]
              valid_mask <- !is.na(h_row) & stringr::str_trim(h_row) != ""
              headers <- ifelse(valid_mask,
                                ifelse(headers == "", h_row, paste0(headers, "_", h_row)),
                                headers)
          }
          headers <- ifelse(headers == "", NA, headers)
      } else {
          headers <- raw_mat[header_row_index, ]
      }
      
      data_block <- raw_mat[true_start:true_end, , drop = FALSE]
      
      block_counts <- num_counts[true_start:true_end]
      internal_valid_rows <- (block_counts >= density_threshold)
      
      # Hierarchical Section Header Propagation
      section_categories <- rep(NA, nrow(data_block))
      has_sections <- FALSE
      
      for (i in seq_len(nrow(data_block))) {
          if (block_counts[i] == 0) {
              row_vals <- data_block[i, ]
              non_empty <- which(!is.na(row_vals) & stringr::str_trim(row_vals) != "")
              if (length(non_empty) == 1 && non_empty[1] <= 3) {
                  section_categories[i] <- stringr::str_trim(row_vals[non_empty[1]])
                  has_sections <- TRUE
              }
          }
      }
      
      if (has_sections) {
          current_section <- NA
          for (i in seq_len(length(section_categories))) {
              if (!is.na(section_categories[i])) {
                  current_section <- section_categories[i]
              } else {
                  section_categories[i] <- current_section
              }
          }
      }
      
      if (has_sections) {
          valid_sections <- section_categories[internal_valid_rows]
      }
      
      data_block <- data_block[internal_valid_rows, , drop = FALSE]
      
      empty_cols <- apply(data_block, 2, function(col) all(is.na(col) | col == "" | col %in% na_strings)) & 
                    (is.na(headers) | headers == "")
                    
      headers <- headers[!empty_cols]
      data_block <- data_block[, !empty_cols, drop = FALSE]
      
      # Intercept and remove repeating page headers
      is_repeated_header <- apply(data_block, 1, function(row) {
          match_count <- sum(stringr::str_trim(row) == stringr::str_trim(headers), na.rm = TRUE)
          match_count >= max(1, length(headers) - 1)
      })
      
      if (any(is_repeated_header)) {
          data_block <- data_block[!is_repeated_header, , drop = FALSE]
          if (has_sections) {
              valid_sections <- valid_sections[!is_repeated_header]
          }
      }
      
      df <- as.data.frame(data_block, stringsAsFactors = FALSE)
      
      if (has_sections) {
          df$section_category <- valid_sections
          headers <- c(headers, "section_category")
      }
      
      # Phase 16: Indentation Hierarchy Extraction
      first_col <- df[[1]]
      if (is.character(first_col)) {
          valid_idx <- which(!is.na(first_col) & first_col != "")
          if (length(valid_idx) > 0) {
              valid_vals <- first_col[valid_idx]
              num_leading_spaces <- nchar(valid_vals) - nchar(stringr::str_trim(valid_vals, side = "left"))
              
              if (max(num_leading_spaces) > 0 && min(num_leading_spaces) == 0 && length(unique(num_leading_spaces)) > 1) {
                  parent_category <- rep(NA, nrow(df))
                  current_parent <- NA
                  
                  for (r in seq_len(nrow(df))) {
                      val <- first_col[r]
                      if (!is.na(val) && val != "") {
                          spaces <- nchar(val) - nchar(stringr::str_trim(val, side = "left"))
                          if (spaces == 0) {
                              current_parent <- stringr::str_trim(val)
                          }
                          parent_category[r] <- current_parent
                      } else {
                          parent_category[r] <- current_parent
                      }
                  }
                  
                  extracted_count <- sum(!is.na(parent_category) & parent_category != stringr::str_trim(first_col), na.rm = TRUE)
                  if (extracted_count > 0) {
                      df$parent_category <- parent_category
                      headers <- c(headers, "parent_category")
                      audit_log[["Indentation Hierarchy Extracted"]] <- extracted_count
                  }
              }
          }
      }
      
      # Phase 6: Forward-fill leading character columns (Staircase Ledgers)
      for (c in 1:min(2, ncol(df))) {
          col_vals <- as.character(df[[c]])
          valid_vals <- col_vals[!is.na(col_vals) & stringr::str_trim(col_vals) != ""]
          if (length(valid_vals) > 0) {
              num_count <- sum(vapply(valid_vals, is_numeric_like, logical(1)))
              if (num_count < length(valid_vals) * 0.5) {
                  filled_col <- col_vals
                  last_val <- NA
                  for (r in seq_along(filled_col)) {
                      if (!is.na(filled_col[r]) && stringr::str_trim(filled_col[r]) != "") {
                          last_val <- filled_col[r]
                      } else if (!is.na(last_val)) {
                          filled_col[r] <- last_val
                      }
                  }
                  df[[c]] <- filled_col
              }
          }
      }
      
      # Phase 5: Amputate Mid-Table Subtotals
      subtotal_keywords <- c("subtotal", "\u5c0f\u8ba1", "total:", "sum:", "gesamt", "summe", "somme", "promedio")
      first_col_lower <- stringr::str_trim(tolower(as.character(df[[1]])))
      subtotal_idx <- vapply(first_col_lower, function(val) {
          if (is.na(val)) return(FALSE)
          any(vapply(subtotal_keywords, function(k) grepl(k, val, fixed = TRUE), logical(1)))
      }, logical(1))
      if (any(subtotal_idx)) {
          audit_log[["Mid-Table Subtotals Amputated"]] <- sum(subtotal_idx)
          df <- df[!subtotal_idx, , drop = FALSE]
          # Recalculate first_col_lower for the next steps
          first_col_lower <- stringr::str_trim(tolower(as.character(df[[1]])))
      }
      
      # Phase 6: Footnote Amputator (Trailing long-string rows)
      tail_n <- min(5, nrow(df))
      footnotes_dropped <- 0
      if (tail_n > 0) {
          rows_to_keep <- rep(TRUE, nrow(df))
          for (r in seq(nrow(df) - tail_n + 1, nrow(df))) {
              val1 <- as.character(df[r, 1])
              if (!is.na(val1) && nchar(stringr::str_trim(val1)) > 15) {
                  other_cols_empty <- TRUE
                  if (ncol(df) > 1) {
                      other_vals <- as.character(df[r, 2:ncol(df)])
                      if (any(!is.na(other_vals) & stringr::str_trim(other_vals) != "")) {
                          other_cols_empty <- FALSE
                      }
                  }
                  if (other_cols_empty) {
                      rows_to_keep[r] <- FALSE
                      footnotes_dropped <- footnotes_dropped + 1
                  }
              }
          }
          df <- df[rows_to_keep, , drop = FALSE]
      }
      audit_log[["Footnotes Dropped"]] <- footnotes_dropped
      
      # Phase 4: Amputate Trailing Aggregation Rows (Ghost Bottoms)
      tail_n <- min(5, nrow(df))
      if (tail_n > 0) {
          agg_keywords <- c("total", "sum", "average", "avg", "\u5408\u8ba1", "\u603b\u8ba1", "\u5e73\u5747", "mean", "gesamt", "summe", "durchschnitt", "moyenne", "somme", "promedio")
          for (r in seq(nrow(df) - tail_n + 1, nrow(df))) {
              if (!is.na(first_col_lower[r]) && any(vapply(agg_keywords, function(k) grepl(paste0("^", k), first_col_lower[r]), logical(1)))) {
                  audit_log[["Ghost Bottom Rows Dropped"]] <- nrow(df) - r + 1
                  if (r > 1) {
                      df <- df[1:(r-1), , drop = FALSE]
                  } else {
                      df <- df[0, , drop = FALSE]
                  }
                  break
              }
          }
      }
      
      # Phase 7: Sanitize raw headers (remove \n and \r)
      headers <- vapply(headers, function(x) {
          if (!is.na(x)) {
              x <- gsub("[\r\n]+", " ", x)
              x <- stringr::str_squish(x)
          }
          x
      }, character(1))
      
      # Phase 11: Orphaned Header Re-Alignment
      if (length(headers) > 0 && (is.na(headers[1]) || headers[1] == "")) {
          headers[1] <- "Category"
      }
      
      colnames(df) <- make.unique(stringr::str_trim(headers), sep = "_")
      
      # Phase 7: Amputate ALL Aggregation Columns (Embedded Subtotals)
      col_agg_keywords <- c("total", "sum", "subtotal", "ytd", "\u5408\u8ba1", "\u603b\u8ba1", "\u5c0f\u8ba1", "average", "avg", "gesamt", "summe", "durchschnitt", "moyenne", "somme", "promedio")
      cols_to_keep <- rep(TRUE, ncol(df))
      subtotal_cols_dropped <- c()
      for (c in seq_len(ncol(df))) {
          col_name <- tolower(colnames(df)[c])
          if (any(vapply(col_agg_keywords, function(k) grepl(k, col_name, fixed = TRUE), logical(1)))) {
              if (c > 1) { # Protect the first column
                  cols_to_keep[c] <- FALSE
                  subtotal_cols_dropped <- c(subtotal_cols_dropped, col_name)
              }
          }
      }
      audit_log[["Subtotal Columns Amputated"]] <- length(subtotal_cols_dropped)
      df <- df[, cols_to_keep, drop = FALSE]
      
      # Phase 11: Phantom Column Purge (Information Density ~ 0)
      phantom_cols <- vapply(df, function(col) {
          valid_vals <- col[!is.na(col) & stringr::str_trim(col) != "" & !(col %in% na_strings)]
          length(valid_vals) == 0
      }, logical(1))
      no_header <- is.na(colnames(df)) | grepl("^(na|x|\\.\\.\\.)[_0-9]*$|^$", tolower(colnames(df)))
      if (any(phantom_cols & no_header)) {
          audit_log[["Phantom Columns Purged"]] <- sum(phantom_cols & no_header)
          df <- df[, !(phantom_cols & no_header), drop = FALSE]
      }
      
      # (section_category assigned earlier)
      
      # Forward-fill NAs in leading character columns (Handling Merged Cells)
      for (j in seq_len(ncol(df))) {
          col_vals <- df[[j]]
          valid_idx <- which(!is.na(col_vals) & col_vals != "")
          valid_count <- length(valid_idx)
          
          if (valid_count > 0) {
              num_likes <- sum(vapply(col_vals[valid_idx], is_numeric_like, logical(1)))
              if ((num_likes / valid_count) < 0.5) {
                  if (sum(is.na(col_vals) | col_vals == "") > 0) {
                      for (r in 2:nrow(df)) {
                          if (is.na(df[r, j]) || df[r, j] == "") {
                              df[r, j] <- df[r - 1, j]
                          }
                      }
                  }
              } else {
                  break
              }
          } else {
              break
          }
      }
      
      df[] <- lapply(df, function(x) {
        x[x %in% na_strings] <- NA
        x
      })
      
      df[] <- lapply(df, function(x) {
        clean_x <- stringr::str_remove_all(x, intToUtf8(160))
        
        # Phase 4: Convert Accounting Zeros to 0
        x_trimmed <- stringr::str_trim(clean_x)
        dash_idx <- x_trimmed %in% c("-", "\u2013", "\u2014")
        if (any(dash_idx, na.rm = TRUE)) {
            clean_x[which(dash_idx)] <- "0"
        }
        
        # Phase 5: Non-standard scientific notation
        clean_x <- stringr::str_replace_all(clean_x, "(?i)\\s*[x\\*]\\s*10\\^([\\-\\+]?[0-9]+)", "E\\1")
        
        clean_x <- stringr::str_replace(clean_x, "^\\s*\\((.*)\\)\\s*$", "-\\1")
        clean_x <- stringr::str_replace(clean_x, "^-\\s+", "-")
        clean_x <- stringr::str_replace(clean_x, "^\\s*([0-9.,\\s]+?)\\s*-$", "-\\1")
        is_pct <- grepl("%\\s*$", clean_x) & !is.na(clean_x)
        
        clean_x <- stringr::str_remove_all(clean_x, "[\\$\u20ac\u00a3\u00a5%\u4e07\u4ebf\u5143\u5343]")
        clean_x <- stringr::str_trim(clean_x)
        
        clean_x <- stringr::str_replace_all(clean_x, "(?<=\\d)[\\s\\u00A0'](?=\\d)", "")
        has_euro_decimal <- grepl(",[0-9]{1,2}[^0-9]*$", clean_x)
        clean_x <- ifelse(has_euro_decimal & !is.na(has_euro_decimal), 
                          stringr::str_replace(stringr::str_remove_all(clean_x, "\\."), ",", "."), 
                          clean_x)
        clean_x <- stringr::str_remove_all(clean_x, ",")
        
        # Phase 11: Semantic Multiplier Engine
        multiplier <- rep(1, length(clean_x))
        k_idx <- grepl("(?i)[-0-9.]+\\s*k$", clean_x) & !is.na(clean_x)
        m_idx <- grepl("(?i)[-0-9.]+\\s*(m|mil|million)$", clean_x) & !is.na(clean_x)
        b_idx <- grepl("(?i)[-0-9.]+\\s*(b|bn|billion)$", clean_x) & !is.na(clean_x)
        t_idx <- grepl("(?i)[-0-9.]+\\s*(t|tn|trillion)$", clean_x) & !is.na(clean_x)
        
        multiplier[k_idx] <- 1000
        multiplier[m_idx] <- 1000000
        multiplier[b_idx] <- 1000000000
        multiplier[t_idx] <- 1000000000000
        
        clean_x <- stringr::str_replace(clean_x, "(?i)\\s*(k|m|mil|million|b|bn|billion|t|tn|trillion)$", "")
        
        clean_x <- stringr::str_replace(clean_x, "\\s*\\*+\\s*$", "")
        clean_x <- stringr::str_replace(clean_x, "\\s*[\\(\\[].*?[\\)\\]]\\s*$", "")
        
        # Exclude quarters from being stripped and converted
        is_quarter <- grepl("(?i)^(Q[1-4]|H[1-2]|FY[0-9]+)$", stringr::str_trim(x))
        
        clean_x[!is_quarter] <- stringr::str_replace(clean_x[!is_quarter], "\\s*[A-Za-z\u4e00-\u9fa5]+\\s*$", "")
        clean_x[!is_quarter] <- stringr::str_replace(clean_x[!is_quarter], "^\\s*[A-Za-z\u4e00-\u9fa5]+\\s*", "")
        
        num_x <- suppressWarnings(as.numeric(clean_x))
        
        if(sum(is.na(num_x)) == sum(is.na(x))) {
            num_x[is_pct & !is.na(num_x)] <- num_x[is_pct & !is.na(num_x)] / 100
            valid_num_idx <- !is.na(num_x)
            num_x[valid_num_idx] <- num_x[valid_num_idx] * multiplier[valid_num_idx]
            return(num_x)
        } 
        
        final_str <- stringr::str_replace(x, "^\\s*[\\.-]+\\s*", "")
        return(stringr::str_trim(final_str))
      })
      
      colnames(df) <- tolower(colnames(df)) 
      if (clean_vars) {
        df <- clean_variable_names(df)
      }
      
      # Phase 14: Auto-Pivot Engine (Wide to Long)
      if (auto_pivot && nrow(df) > 0) {
          cnames <- colnames(df)
          temporal_pattern <- "^(19|20)[0-9]{2}(_q[1-4]|_h[1-2]|_[0-1]?[0-9]|_[a-z]{3})?$|^(q[1-4]|h[1-2]|fy[0-9]+)$|^[a-z]{3}_([0-9]{2}|(19|20)[0-9]{2})$"
          is_temporal <- grepl(temporal_pattern, cnames)
          
          if (sum(is_temporal) >= 2) {
              id_cols <- cnames[!is_temporal]
              temporal_cols <- cnames[is_temporal]
              
              # Base R melt implementation
              long_list <- lapply(temporal_cols, function(tc) {
                  sub_df <- df[, id_cols, drop = FALSE]
                  sub_df$time_period <- tc
                  sub_df$value <- df[[tc]]
                  return(sub_df)
              })
              
              df <- do.call(rbind, long_list)
              rownames(df) <- NULL
              audit_log[["Auto-Pivot Wide to Long"]] <- sum(is_temporal)
          }
      }
      
      list(success = TRUE, data = df, audit = audit_log)
    }, error = function(e) {
      list(success = FALSE, error = e$message)
    })
    
    if (res$success) {
      if (return_audit) {
          audit_df <- data.frame(
              Operation = names(res$audit),
              Count = as.character(unlist(res$audit, use.names = FALSE)),
              stringsAsFactors = FALSE
          )
          audit_df <- audit_df[audit_df$Count != "0", , drop = FALSE]
          rownames(audit_df) <- NULL
          return(list(data = res$data, audit = audit_df))
      }
      return(res$data)
    } else {
      last_error <- res$error
    }
  }
  
  stop(paste("Failed to parse any valid panel from file.", last_error))
}

test_that("Engine parses a standard clean panel", {
    tmp <- tempfile(fileext = ".xlsx")
    df <- data.frame(
        Category = c("A", "B"),
        Value = c("100", "200"),
        stringsAsFactors = FALSE
    )
    writexl::write_xlsx(df, tmp)

    res <- read_messy_panel(tmp)
    expect_equal(nrow(res), 2)
    expect_equal(as.numeric(res$value), c(100, 200))
    unlink(tmp)
})

#' Standardize and Clean Variable Names
#'
#' @description
#' `clean_variable_names()` standardizes column names in a messy data frame. It converts all names 
#' to snake_case, strips special characters (except `_`), translates Excel serial dates (e.g., `44197`) 
#' into ISO date strings (`2021-01-01`), and maps common financial/academic synonyms (e.g., `gvkey`, 
#' `permno`, `cusip`) to standard names (`id`, `ticker`).
#'
#' @param data A `data.frame`. The data frame with messy column names.
#' @return A `data.frame` with the same data but standardized column names.
#' 
#' @examples
#' # Toy example: standardize column names in a data frame
#' df <- data.frame(
#'   `Total Revenue ($)` = 1,
#'   `PERMNO` = 3,
#'   `My Custom Column!` = 4,
#'   check.names = FALSE
#' )
#' clean_df <- clean_variable_names(df)
#' colnames(clean_df)
#' # Returns: c("revenue", "id", "my_custom_column")
#'
#' # Excel serial dates are also handled
#' df2 <- data.frame(`44197` = 2, check.names = FALSE)
#' colnames(clean_variable_names(df2))
#' # Returns: "2021-01-01"
#'
#' @export
#' @importFrom stringr str_remove_all str_trim str_to_lower
clean_variable_names <- function(data) {
  clean_names <- stringr::str_trim(colnames(data))
  clean_names <- stringr::str_to_lower(clean_names)
  
  # Check if the name is an Excel serial date (e.g. 44197 -> 2021-01-01)
  is_excel_date <- grepl("^[345][0-9]{4}$", clean_names)
  if (any(is_excel_date)) {
      clean_names[is_excel_date] <- as.character(as.Date(as.numeric(clean_names[is_excel_date]), origin = "1899-12-30"))
  }
  
  dict <- c(
    "gvkey" = "id",
    "permno" = "id",
    "global company key" = "id",
    "company id" = "id",
    "entity id" = "id",
    "patient mrn" = "id",
    "provider id" = "id",
    "employee no." = "id",
    "tracking id" = "id",
    "\u8eab\u4efd\u8bc1\u53f7" = "id",
    "patienten-id" = "id",
    "num\u00e9ro de patient" = "id",
    
    "datadate" = "date",
    "date" = "date",
    "data date" = "date",
    "fiscal year" = "date",
    "report date" = "date",
    "admission date" = "date",
    "pay period" = "date",
    "posting date" = "date",
    "dispatch date" = "date",
    "\u65e5\u671f" = "date",
    "datum" = "date",
    "fecha" = "date",
    
    "at" = "total_assets",
    "assets - total" = "total_assets",
    "assets total" = "total_assets",
    
    "lt" = "total_liabilities",
    "liabilities - total" = "total_liabilities",
    "liabilities total" = "total_liabilities",
    
    "sic" = "category",
    "standard industry classification code" = "category",
    "industry code" = "category",
    "sector" = "category",
    "icd-10 code" = "category",
    "department" = "category",
    "cost center" = "category",
    "g/l account" = "category",
    "destination" = "category",
    "state" = "category",
    "soc code" = "category",
    "\u7c7b\u522b" = "category",
    "kategorie" = "category",
    "cat\u00e9gorie" = "category",
    "categor\u00eda" = "category",
    
    "conm" = "name",
    "company name" = "name",
    "ticker" = "name",
    "ticker symbol" = "name",
    "hospital name" = "name",
    "\u540d\u79f0" = "name",
    "nom" = "name",
    "nombre" = "name",
    
    "billing amount" = "value",
    "insurance copay" = "value",
    "total charges" = "value",
    "net revenue" = "value",
    "hourly rate" = "value",
    "amount (usd)" = "value",
    "shipping cost" = "value",
    "mean hourly wage" = "value",
    "annual mean wage" = "value",
    "\u91d1\u989d" = "value",
    "betrag" = "value",
    "montant" = "value",
    "importe" = "value",
    
    "document no." = "ref",
    "reference" = "ref"
  )
  
  clean_names <- stringr::str_remove_all(clean_names, "\\s*[a-z]/\\s*$")
  clean_names <- stringr::str_remove_all(clean_names, "\\s*\\([^)]*\\)\\s*$")
  clean_names <- stringr::str_trim(clean_names)
  
  regex_dict <- list(
      "revenue" = c("revenue", "sales", "turnover", "umsatz", "chiffre d'affaires", "ingresos"),
      "profit" = c("profit", "margin", "income", "gewinn", "b\u00e9n\u00e9fice", "beneficio"),
      "cost" = c("cost", "expense", "cogs", "kosten", "d\u00e9pense", "gasto")
  )
  
  for (i in seq_along(clean_names)) {
    matched <- FALSE
    
    # 1. Exact Match
    if (clean_names[i] %in% names(dict)) {
      clean_names[i] <- dict[[clean_names[i]]]
      matched <- TRUE
    }
    
    # 2. Regex Fuzzy Match
    if (!matched) {
      for (target in names(regex_dict)) {
        patterns <- regex_dict[[target]]
        if (any(vapply(patterns, function(p) grepl(p, clean_names[i], ignore.case = TRUE), logical(1)))) {
            clean_names[i] <- target
            matched <- TRUE
            break
        }
      }
    }
    
    # 2.5 ISO Date Pass-through
    if (!matched && grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2}$", clean_names[i])) {
        matched <- TRUE
    }
    
    # 3. Strict snake_case conversion for unmapped variables
    if (!matched) {
        snaked <- stringr::str_replace_all(clean_names[i], "[^a-z0-9_]+", "_")
        snaked <- stringr::str_replace_all(snaked, "_+", "_")
        snaked <- stringr::str_replace(snaked, "^_|_$", "")
        if (snaked != "") {
            clean_names[i] <- snaked
        }
    }
  }
  
  colnames(data) <- clean_names
  return(data)
}

test_that("clean_variable_names maps synonyms and enforces snake_case", {
    df <- data.frame(
        `GVKEY` = 1,
        `Total Revenue ($)` = 2,
        `My Custom Column` = 3,
        check.names = FALSE
    )
    res <- clean_variable_names(df)
    expect_equal(colnames(res), c("id", "revenue", "my_custom_column"))
})

#' Smart Type Coercion & NA Recognition
#' 
#' @description
#' `infer_data_types()` scans character columns in a data frame, identifies common
#' financial placeholders for missing data (e.g., "-", "N/A", "n.m."), safely replaces 
#' them with `NA`, and then coerces the column to `numeric` or `Date` if a high 
#' percentage of the remaining values match those types.
#' 
#' @param data A `data.frame`.
#' @param na_strings A character vector of strings to be interpreted as `NA`.
#' @param num_threshold Numeric between 0 and 1. The proportion of valid numbers required to convert a column to numeric. Default is `0.95`.
#' @return A `data.frame` with inferred data types.
#' 
#' @examples
#' # Clean financial placeholders and coerce to numeric
#' df <- data.frame(val = c("1.5", "-", "2.0", "N/A"), stringsAsFactors = FALSE)
#' df_clean <- infer_data_types(df)
#' df_clean$val  # numeric: c(1.5, NA, 2.0, NA)
#' is.numeric(df_clean$val)  # TRUE
#'
#' @export
#' @importFrom stringr str_trim
infer_data_types <- function(data, na_strings = c("-", "N/A", "n/a", "n.m.", "n.m", "NA", "null", "NULL", "."), num_threshold = 0.95) {
    
    for (i in seq_along(data)) {
        if (is.character(data[[i]])) {
            col_data <- stringr::str_trim(data[[i]])
            
            # Replace defined NA strings with actual NA
            is_na_string <- tolower(col_data) %in% tolower(na_strings)
            col_data[is_na_string] <- NA
            
            # Count valid elements
            valid_elements <- col_data[!is.na(col_data) & col_data != ""]
            
            if (length(valid_elements) > 0) {
                # Test for numeric
                num_vals <- suppressWarnings(as.numeric(valid_elements))
                num_ratio <- sum(!is.na(num_vals)) / length(valid_elements)
                
                if (num_ratio >= num_threshold) {
                    # Safe to convert to numeric
                    col_data[col_data == ""] <- NA
                    data[[i]] <- suppressWarnings(as.numeric(col_data))
                    next
                }
                
                # Test for Dates
                date_vals <- suppressWarnings(as.Date(valid_elements, format = "%Y-%m-%d"))
                if (sum(!is.na(date_vals)) / length(valid_elements) >= num_threshold) {
                    col_data[col_data == ""] <- NA
                    data[[i]] <- suppressWarnings(as.Date(col_data, format = "%Y-%m-%d"))
                    next
                }
                
                # Alternate date formats can be added here
                date_vals2 <- suppressWarnings(as.Date(valid_elements, format = "%Y/%m/%d"))
                if (sum(!is.na(date_vals2)) / length(valid_elements) >= num_threshold) {
                    col_data[col_data == ""] <- NA
                    data[[i]] <- suppressWarnings(as.Date(col_data, format = "%Y/%m/%d"))
                    next
                }
                
                # If neither numeric nor date ratio met the threshold, assign NA-cleaned vector
                data[[i]] <- col_data
            } else {
                # Entire column is NA or empty
                col_data[col_data == ""] <- NA
                # Convert fully empty character columns to logical NAs to match standard read_csv behavior
                data[[i]] <- as.logical(col_data) 
            }
        }
    }
    
    return(data)
}

test_that("infer_data_types cleans financial NAs and coerces to numeric", {
    df <- data.frame(
        val = c("1.5", "-", "2.0", "N/A"),
        stringsAsFactors = FALSE
    )
    res <- infer_data_types(df)
    expect_true(is.numeric(res$val))
    expect_equal(res$val, c(1.5, NA, 2.0, NA))
})

#' Normalize Numeric Columns Based on Header Unit Declarations
#' 
#' @description
#' `normalize_units()` scans the column names of a data frame for financial/scientific 
#' unit declarations (e.g., "Revenue (in millions)", "Assets ($k)", "Employees ('000)").
#' It automatically multiplies the numeric values in the corresponding columns by the 
#' detected multiplier (1,000, 1,000,000, etc.) and optionally strips the unit 
#' declaration from the column name.
#' 
#' @param data A `data.frame`. The data frame to be normalized.
#' @param strip_units Logical. If `TRUE`, removes the unit declarations from the column names. Default is `TRUE`.
#' @return A `data.frame` with the normalized data and updated column names.
#' 
#' @examples
#' # Scale columns declared in millions and thousands
#' df <- data.frame(
#'   `Revenue ($M)` = c(1.5, 2.0),
#'   `Cost (in thousands)` = c(500, 600),
#'   check.names = FALSE
#' )
#' result <- normalize_units(df)
#' result$Revenue  # c(1500000, 2000000)
#' result$Cost     # c(500000, 600000)
#'
#' @export
#' @importFrom stringr str_remove str_replace str_trim
normalize_units <- function(data, strip_units = TRUE) {
    cnames <- colnames(data)
    
    # Define regex patterns for different units
    p_k <- "(?i)(\\bin thousands\\b|\\bin 000s\\b|'000s|'000|\\$k\\b|\\$ k\\b|\\(\\s*k\\s*\\)|\\(\u5343\\)|\uff08\u5343\uff09)"
    p_m <- "(?i)(\\bin millions\\b|\\bin mln\\b|\\bin millionen\\b|\\ben millions\\b|\\$m\\b|\\$ m\\b|\\(\\s*m\\s*\\)|\\(\u767e\u4e07\\)|\uff08\u767e\u4e07\uff09)"
    p_b <- "(?i)(\\bin billions\\b|\\bin bln\\b|\\bin bn\\b|\\$b\\b|\\$ b\\b|\\(\\s*b\\s*\\)|\\(\u5341\u4ebf\\)|\uff08\u5341\u4ebf\uff09)"
    p_t <- "(?i)(\\bin trillions\\b|\\$t\\b|\\$ t\\b|\\(\\s*t\\s*\\)|\\(\u5146\\)|\uff08\u5146\uff09)"
    
    for (i in seq_along(cnames)) {
        col_name <- cnames[i]
        multiplier <- 1
        matched_pattern <- NULL
        
        if (grepl(p_t, col_name)) {
            multiplier <- 1e12
            matched_pattern <- p_t
        } else if (grepl(p_b, col_name)) {
            multiplier <- 1e9
            matched_pattern <- p_b
        } else if (grepl(p_m, col_name)) {
            multiplier <- 1e6
            matched_pattern <- p_m
        } else if (grepl(p_k, col_name)) {
            multiplier <- 1e3
            matched_pattern <- p_k
        }
        
        if (multiplier > 1) {
            if (is.numeric(data[[i]])) {
                data[[i]] <- data[[i]] * multiplier
            } else {
                num_vals <- suppressWarnings(as.numeric(as.character(data[[i]])))
                if (!all(is.na(num_vals[!is.na(data[[i]])]))) {
                     is_num_idx <- !is.na(num_vals)
                     # Convert character to numeric for rows where it's valid
                     data[[i]] <- as.character(data[[i]])
                     data[[i]][is_num_idx] <- as.character(num_vals[is_num_idx] * multiplier)
                }
            }
            
            if (strip_units) {
                new_name <- stringr::str_remove(col_name, matched_pattern)
                new_name <- stringr::str_replace(new_name, "\\(\\s*\\)|\uff08\\s*\uff09", "")
                new_name <- stringr::str_trim(new_name)
                cnames[i] <- new_name
            }
        }
    }
    
    if (strip_units) {
        colnames(data) <- cnames
    }
    
    return(data)
}

test_that("normalize_units scales columns declared in millions", {
    df <- data.frame(
        `Revenue ($M)` = c(1.5, 2.0),
        check.names = FALSE
    )
    res <- normalize_units(df)
    expect_equal(res$`Revenue`, c(1500000, 2000000))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/infer_data_types.R
\name{infer_data_types}
\alias{infer_data_types}
\title{Smart Type Coercion & NA Recognition}
\usage{
infer_data_types(
  data,
  na_strings = c("-", "N/A", "n/a", "n.m.", "n.m", "NA", "null", "NULL", "."),
  num_threshold = 0.95
)
}
\arguments{
\item{data}{A \code{data.frame}.}

\item{na_strings}{A character vector of strings to be interpreted as \code{NA}.}

\item{num_threshold}{Numeric between 0 and 1. The proportion of valid numbers required to convert a column to numeric. Default is \code{0.95}.}
}
\value{
A \code{data.frame} with inferred data types.
}
\description{
\code{infer_data_types()} scans character columns in a data frame, identifies common
financial placeholders for missing data (e.g., "-", "N/A", "n.m."), safely replaces
them with \code{NA}, and then coerces the column to \code{numeric} or \code{Date} if a high
percentage of the remaining values match those types.
}
\examples{
# Clean financial placeholders and coerce to numeric
df <- data.frame(val = c("1.5", "-", "2.0", "N/A"), stringsAsFactors = FALSE)
df_clean <- infer_data_types(df)
df_clean$val  # numeric: c(1.5, NA, 2.0, NA)
is.numeric(df_clean$val)  # TRUE

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_variable_names.R
\name{clean_variable_names}
\alias{clean_variable_names}
\title{Standardize and Clean Variable Names}
\usage{
clean_variable_names(data)
}
\arguments{
\item{data}{A \code{data.frame}. The data frame with messy column names.}
}
\value{
A \code{data.frame} with the same data but standardized column names.
}
\description{
\code{clean_variable_names()} standardizes column names in a messy data frame. It converts all names
to snake_case, strips special characters (except \verb{_}), translates Excel serial dates (e.g., \code{44197})
into ISO date strings (\code{2021-01-01}), and maps common financial/academic synonyms (e.g., \code{gvkey},
\code{permno}, \code{cusip}) to standard names (\code{id}, \code{ticker}).
}
\examples{
# Toy example: standardize column names in a data frame
df <- data.frame(
  `Total Revenue ($)` = 1,
  `PERMNO` = 3,
  `My Custom Column!` = 4,
  check.names = FALSE
)
clean_df <- clean_variable_names(df)
colnames(clean_df)
# Returns: c("revenue", "id", "my_custom_column")

# Excel serial dates are also handled
df2 <- data.frame(`44197` = 2, check.names = FALSE)
colnames(clean_variable_names(df2))
# Returns: "2021-01-01"

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_messy_panel.R
\name{read_messy_panel}
\alias{read_messy_panel}
\title{Robust Parsing and Extraction of Messy Excel Panel Data}
\usage{
read_messy_panel(
  file_path,
  sheet = NULL,
  na_strings = c("", "NA", "#N/A", "NULL", "S", "D", "ND", "N/A", "*", "**", "***", ".",
    "x", "c", "s", "z", "#VALUE!", "#REF!", "#DIV/0!", "#NUM!", "#NAME?", "none", "NR",
    "--", "---", "n.a.", "N.A.", "n/a", "Not Applicable"),
  clean_vars = TRUE,
  auto_pivot = FALSE,
  return_audit = FALSE
)
}
\arguments{
\item{file_path}{Character string. Path to the Excel file.}

\item{sheet}{Optional sheet name or index. If \code{NULL} (the default), it auto-discovers the first valid data panel across all sheets. If \code{"ALL"}, parses and merges all sheets.}

\item{na_strings}{Character vector. Strings to interpret as missing values. Supports complex missing-value lexicons.}

\item{clean_vars}{Logical. If \code{TRUE} (default), standardizes variable names to snake_case using \code{clean_variable_names()}.}

\item{auto_pivot}{Logical. If \code{TRUE}, attempts to reshape wide temporal columns (e.g., FY2021, Q1_2022) into a long format (\code{time_period}, \code{value}).}

\item{return_audit}{Logical. If \code{TRUE}, returns a list containing \verb{$data} (the cleaned data frame) and \verb{$audit} (a detailed log of all algorithmic modifications made).}
}
\value{
If \code{return_audit = FALSE}, a cleaned and standardized \code{data.frame}.
If \code{return_audit = TRUE}, a named list containing:
\item{data}{The cleaned \code{data.frame}.}
\item{audit}{A \code{data.frame} detailing exactly what transformations, truncations, or imputations were applied.}
}
\description{
\code{read_messy_panel()} is an industrial-grade parser designed to extract clean, standardized data frames
from heavily malformed, human-readable Excel reports (e.g., financial statements, ERP exports).
It automatically bypasses decoy rows, stitches N-dimensional hierarchical headers, extracts structural
indentation hierarchies (parent-child relationships), amputates embedded subtotals, and standardizes
financial/scientific numbers.
}
\examples{
# Toy example: create a small in-memory Excel file and parse it
tmp <- tempfile(fileext = ".xlsx")
df_raw <- data.frame(
  Category = c("Revenue", "Cost", "Total"),
  `2022` = c("1.2M", "800k", "2.0M"),
  `2023` = c("1.5M", "900k", "2.4M"),
  check.names = FALSE
)
writexl::write_xlsx(df_raw, tmp)
result <- read_messy_panel(tmp, auto_pivot = TRUE)
head(result)
unlink(tmp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalize_units.R
\name{normalize_units}
\alias{normalize_units}
\title{Normalize Numeric Columns Based on Header Unit Declarations}
\usage{
normalize_units(data, strip_units = TRUE)
}
\arguments{
\item{data}{A \code{data.frame}. The data frame to be normalized.}

\item{strip_units}{Logical. If \code{TRUE}, removes the unit declarations from the column names. Default is \code{TRUE}.}
}
\value{
A \code{data.frame} with the normalized data and updated column names.
}
\description{
\code{normalize_units()} scans the column names of a data frame for financial/scientific
unit declarations (e.g., "Revenue (in millions)", "Assets ($k)", "Employees ('000)").
It automatically multiplies the numeric values in the corresponding columns by the
detected multiplier (1,000, 1,000,000, etc.) and optionally strips the unit
declaration from the column name.
}
\examples{
# Scale columns declared in millions and thousands
df <- data.frame(
  `Revenue ($M)` = c(1.5, 2.0),
  `Cost (in thousands)` = c(500, 600),
  check.names = FALSE
)
result <- normalize_units(df)
result$Revenue  # c(1500000, 2000000)
result$Cost     # c(500000, 600000)

}
